#!/usr/bin/python

import os
import subprocess
import re

raise ValueError("MLHP_ENABLE_OMP changed to MLHP_MULTITHREADING=OMP")

# (polynomial degree, number of elements, refinement level)
configurations = [(1, 140, 0), (2, 65, 0), (3, 34, 0), (5, 12, 0), (8, 4, 0), (3, 8, 3)]

#flags = ("-march=native -mprefer-vector-width=512", )
flags = ("-march=native -mno-avx512f -mno-avx", "-march=native -mno-avx512f", "-march=native -mprefer-vector-width=512")

#alignments = (64, ) 
alignments = (8, 32, 64)

compileCores = 20

runTimes = 2

numberOfIterations = 50

logFile = "./performance_benchmark"

tmpdir = "tmp_mlhp_benchmark"
source = tmpdir + "/source"
build = tmpdir + "/build"

keywords = ["grid", "basis", "allocation", "shapes", "matrix", "scatter", "assembly", "total"]

if not os.path.exists( tmpdir ):
    os.makedirs( tmpdir )

os.system( "rm -rf " + tmpdir + "/*" )
os.system( "git clone --recursive https://gitlab.com/phmkopp/mlhp.git " + source )
os.system( "cd " + source + "&& git checkout refactorSpatialFunctions && cd -" )

os.system( "mkdir " + build )

def readFile( filename ):
    with open( filename, 'r') as content_file:
        code = content_file.read()
    return code

def writeFile( filename, content ):
    with open( filename, "w" ) as text_file:
        text_file.write( content )

def placeFile( polynomialDegree, numberOfElements, refinementLevel, MatrixType ):
    code = readFile( "benchmark.cpp.in" )
    code = code.replace( '__polynomialDegree__', str( polynomialDegree ) )
    code = code.replace( '__numberOfElements__', str( numberOfElements ) )
    code = code.replace( '__refinementLevel__', str( refinementLevel ) )
    code = code.replace( '__MatrixType__', MatrixType )

    writeFile( os.path.join( source, "tests", "system", "files.cmake" ), "set( MLHP_SYSTEM_TEST_SOURCES benchmark.cpp )" )
    writeFile( os.path.join( source, "tests", "system", "benchmark.cpp" ), code )


def cmakeConfigure( flags, alignment ):
    allFlags = "-Ofast -Wno-unused-parameter -Wno-unused-variable -Wno-unused-local-typedefs" + " " + flags
    arguments = "-DMLHP_ALL_OPTIMIZATIONS=OFF "\
                "-DMLHP_DEBUG_CHECKS=OFF "     \
                "-DMLHP_ENABLE_OMP=OFF "

    arguments += "-DMLHP_SIMD_ALIGNMENT=" + str( alignment ) + " "
    arguments += "-DCMAKE_CXX_FLAGS=\"" + allFlags + "\""

    os.system("cd " + build + " && cmake " + arguments + " ../source"  )

def compileAndRun( ):
    os.system("cd " + build + " && make -j" + str( compileCores ) + " system_testrunner > /dev/null" )

    times = [0.0 for _ in keywords]

    for _ in range( runTimes ):
        os.system(build + "/bin/system_testrunner > " + tmpdir + "/current_benchmark_output.txt" )

        content = readFile( tmpdir + "/current_benchmark_output.txt" )

        for iTime, name in enumerate( keywords ):
            times[iTime] += float( re.search( r"" + name + ": (.*)", content ).group(1) )

    return [time / float( runTimes ) for time in times]

def replaceInFile( filename, old, new ):
    content = readFile( filename )
    content = content.replace( old, new )
    writeFile( filename, content )

writeFile( logFile + "_log.txt", "" )
def log( stuff ):
    print( stuff )
    with open( logFile + "_log.txt","a") as f:
        f.write( stuff + "\n" )

# Inject timing couts into assembly.cpp
def insertLines( filename, lines ):
    with open( filename, "r" ) as file:
        content = file.readlines( )

    for index, (number, line) in enumerate( lines ):
        content.insert( number + index, line + "\n" )

    with open( filename, "w" ) as file:
        for line in content:
            file.write( line );

cppCast = lambda var : "std::chrono::duration_cast<std::chrono::duration<double>>(" + var + " - t0).count( )"
cppPoint = lambda index : "auto t" + str( index ) + " = std::chrono::steady_clock::now( );"
cppCout = lambda text : "std::cout << " + text + " << std::endl;\n"

start = 234

cppCode = [ #( 14, "#include <iostream>"),
            ( start + 22, "auto t0 = std::chrono::steady_clock::now( );\n"\
                          "auto shapesTime  = t0;\n"\
                          "auto matrixTime  = t0;\n"\
                          "auto scatterTime = t0;"),
            ( start + 30, cppPoint(1) ),
            ( start + 31, cppPoint(2) ),
            ( start + 39, cppPoint(3) ),
            ( start + 40, cppPoint(4) ),
            ( start + 40, "shapesTime  += t4 - t3;\n" ),
            ( start + 43, cppPoint(5) ),
            ( start + 44, cppPoint(6) ),
            ( start + 45, cppPoint(7) ),
            ( start + 45, "shapesTime  += t6 - t5;\n"\
                          "matrixTime  += t7 - t6;\n"),
            ( start + 48, cppPoint(8) ),
            ( start + 49, cppPoint(9) ),
            ( start + 49, "shapesTime  += t2 - t1;\n"\
                          "scatterTime += t9 - t8;\n"),
            ( start + 50, cppCout( "\"shapes:  \" << " + cppCast( "shapesTime"  ) ) + \
                          cppCout( "\"matrix:  \" << " + cppCast( "matrixTime"  ) ) + \
                          cppCout( "\"scatter: \" << " + cppCast( "scatterTime" ) ) ) ]
                    
insertLines( source + "/src/core/assembly.cpp", cppCode )
 

# Start measurements
results = [ [ [ [ [ ] for _3 in range( 2 * len( flags ) ) ] for _2 in configurations ] for _1 in alignments ] for _ in keywords ]

def printResults( ):
    for iTiming, name in enumerate( keywords ):
        log( "-------------------- " + name + " time -----------------\n" )
        for iAlign, alignment in enumerate( alignments ):
            resultString = "alignment: " + str( alignment ) + "\n"
            for iConfig, conf in enumerate( configurations ):
                for iCompile, time in enumerate( results[iTiming][iAlign][iConfig] ):
                    resultString += str( sum( time ) / len( time ) ) + ", "
                resultString += "\n"
            log( resultString )

for iteration in range( numberOfIterations ):
    log("============================ iteration " + str( iteration ) + " =================================")
    for iAlign, alignment in enumerate( alignments ):
        for iFlag, flag in enumerate( flags ):
            log( "configuring with " + str( alignment ) + " byte alignement and compiler flags: " + flag )

            placeFile( *configurations[0], MatrixType="Symmetric" )

            cmakeConfigure( flag, alignment )

            for iConfig, config in enumerate( configurations ):
 
               for iSymm, symmetry in enumerate(("Unsymmetric", "Symmetric")):
                    placeFile( *config, MatrixType=symmetry )

                    if symmetry == "Symmetric":
                        replaceInFile( source + "/src/core/integrands.cpp", "linalg::unsymmetricElementLhs", "linalg::symmetricElementLhs" )
                        replaceInFile( source + "/src/core/integrands.cpp", "AssemblyType::UnsymmetricMatrix", "AssemblyType::SymmetricMatrix" )
                    else: 
                        replaceInFile( source + "/src/core/integrands.cpp", "linalg::symmetricElementLhs", "linalg::unsymmetricElementLhs" )
                        replaceInFile( source + "/src/core/integrands.cpp", "AssemblyType::SymmetricMatrix", "AssemblyType::UnsymmetricMatrix" )

                    time = compileAndRun( )
 
                    log( "p = " + str( config[0] ) + ", n = " + str( config[1] ) + ", r = " + str( config[2] ) + ", " + symmetry + ": " + str( time ) )
 
                    for i in range( len( keywords ) ):
                        results[i][iAlign][iConfig][iSymm * len(flags) + iFlag].append( time[i] )

    printResults( )
    writeFile( logFile + "_data.py", "data = " + str( results ) )

os.system( "rm -rf " + tmpdir )



# This file is part of the mlhp project. License: See LICENSE

import os, sys, ast
from functools import reduce

try:
    # mlhp.py script folder
    path = os.path.abspath( os.path.dirname(sys.argv[0]) );
    
    # Try to open path/mlhpPythonPath containing the python module path. This
    # file is written as post build command after compiling pymlhpcore.
    with open( os.path.join( path, 'mlhpPythonPath' ), 'r') as f:
        sys.path.append( os.path.normpath( f.read( ).splitlines( )[0] ) )
   
except IOError: 
    pass

from pymlhpcore import *
from pymlhpcore import _scalarFieldFromTree, _vectorFieldFromTree, _vectorFieldFromAddress, _scalarFieldFromAddress, _domainIntegrandFromAddress

Triangulation = Triangulation3D
TriangleCellAssociation = TriangleCellAssociation3D

def _iterativeSolve(internalSolve, A, b, x0, *, rtol, atol, maxiter, M, residualNorms):
    maxiter_ = len(b) if maxiter is None else maxiter
    
    mA = linearOperator(A) if isinstance(A, AbsSparseMatrix) else A
    mM = noPreconditioner() if M is None else (linearOperator(M) if isinstance(M, AbsSparseMatrix) else M)
    
    solution = DoubleVector(len(b), 0.0) if x0 is None else copy(vector=x0)
    norms = internalSolve(mA, b, x0=solution, rtol=rtol, atol=atol, maxiter=maxiter_, M=mM);
    
    return [solution, norms] if residualNorms else solution
    
def cg(A, b, x0=None, *, rtol=1e-10, atol=0.0, maxiter=None, M=None, residualNorms=False):
    return _iterativeSolve(internalCG, A, b, x0, rtol=rtol, atol=atol, maxiter=maxiter, M=M, residualNorms=residualNorms)

def bicgstab(A, b, x0=None, *, rtol=1e-10, atol=0.0, maxiter=None, M=None, residualNorms=False):
    return _iterativeSolve(internalBiCGStab, A, b, x0, rtol=rtol, atol=atol, maxiter=maxiter, M=M, residualNorms=residualNorms)

def _makeSolve(solver, *,rtol, atol, maxiter):
    def solve(A, b):
        return solver(A, b, rtol=solve.rtol, atol=solve.atol, maxiter=solve.maxiter, M=diagonalPreconditioner(A))
    solve.rtol = rtol
    solve.atol = atol
    solve.maxiter = maxiter
    return solve
    
def makeCGSolver(*, rtol=1e-10, atol=0.0, maxiter=None):
    return _makeSolve(cg, rtol=rtol, atol=atol, maxiter=maxiter)

def makeBiCGStabSolver(*, rtol=1e-10, atol=0.0, maxiter=None):
    return _makeSolve(bicgstab, rtol=rtol, atol=atol, maxiter=maxiter)

def makeScalars(n, value=0.0):
    return [ScalarDouble( value ) for _ in range( n )]
    
def writeBasisOutput(basis, postmesh=None, writer=VtuOutput("output.vtu"), processors=[]):
    kwargs = {'basis': basis, 'writer' : writer if isinstance(writer, MeshWriter) else writer.meshWriter()}
    
    if postmesh is not None:
        kwargs['postmesh'] = postmesh
    if len(processors) > 0:
        convert = lambda p : type(p).__name__[:-2] != 'ElementProcessor'
        kwargs['processors'] = [(convertToElementProcessor(p) if convert(p) else p) for p in processors]
            
    internalWriteBasisOutput(**kwargs)
 
def writeMeshOutput(mesh, postmesh=None, writer=VtuOutput("output.vtu"), processors=[]):
    kwargs = {'mesh': mesh, 'writer' : writer if isinstance(writer, MeshWriter) else writer.meshWriter()}
    
    if postmesh is not None:
        kwargs['postmesh'] = postmesh
    if len(processors) > 0:
        kwargs['processors'] = processors
            
    internalWriteMeshOutput(**kwargs)
 
def _parseFunction(tree):
    tokens = []
    
    def _convert(node):
        id = _convert.index
        _convert.index += 1
        if isinstance(node, ast.AST):
            nodeType = node.__class__.__name__
            if nodeType == "Constant":
                tokens.append([id, nodeType, str(node.value)])
            elif nodeType == "BinOp":
                tokens.append([id, nodeType, node.op.__class__.__name__, _convert(node.left), _convert(node.right)])
            elif nodeType == "BoolOp" and len(node.values) == 2:
                tokens.append([id, nodeType, node.op.__class__.__name__, _convert(node.values[0]), _convert(node.values[1])])
            elif nodeType == "Compare" and len(node.comparators) == 1:
                tokens.append([id, nodeType, node.ops[0].__class__.__name__, _convert(node.left), _convert(node.comparators[0])])
            elif nodeType == "Call":
                tokens.append([id, nodeType, node.func.id] + [_convert(arg) for arg in node.args])
            elif nodeType == "Name" and node.id in {'x', 'y', 'z', 'r', 's', 't'}:
                tokens.append([id, "Input", str( { 'x' : 0, 'y' : 1, 'z' : 2, 'r' : 0, 's' : 1, 't' : 2 }[node.id] ) ])
            elif nodeType == "Subscript" and isinstance(node.slice, ast.Constant):
                tokens.append([id, "Input", str(node.slice.value)])
            elif nodeType == "UnaryOp":
                tokens.append([id, nodeType, node.op.__class__.__name__, _convert(node.operand)])
            elif nodeType == "IfExp":
                tokens.append([id, "Call", "select", _convert(node.test), _convert(node.body), _convert(node.orelse)])
            elif nodeType == "Num": # Legacy python 3.7
                tokens.append([id, "Constant", str(node.n)])
            elif nodeType == "Subscript" and node.slice.__class__.__name__ == "Index": # Legacy python 3.7
                tokens.append([id, "Input", str(node.slice.value.n)])
            else:
                raise(ValueError("Expression of type \"" + nodeType + "\" is not supported."))
        return str(id)
        
    _convert.index = 0
    _convert(tree)
    
    return [token[1:] for token in sorted(tokens, key=lambda token : token[0])]
    
def _parseScalarField(expr):
    if not isinstance(expr, str): raise ValueError("Expression must be a string")
    
    tree = ast.parse(expr).body
    
    if len(tree) != 1: raise ValueError("Expression string must contain one expression.")
    
    return _parseFunction(tree[0].value)
     
def _parseVectorField(expr):
    if not isinstance(expr, str): raise ValueError("Expression must be a string")
    
    expressionList = ast.parse(expr).body[0].value
    
    if not isinstance(expressionList, ast.List): 
        raise ValueError("Root expression is not a list.")
    
    return [_parseFunction(expr) for expr in expressionList.elts]
    
def scalarField(ndim, func=None, address=None):
    if address is not None:
        if func is not None: raise ValueError("Both function and address given.")
        return _scalarFieldFromAddress(ndim, address)
    if hasattr(func, "address"):
        return _scalarFieldFromAddress(ndim, func.address)
    if isinstance(func, (bool, int, float)):
        return _scalarFieldFromTree(ndim, _parseScalarField(str(float(func))))
    if isinstance(func, str):
        return _scalarFieldFromTree(ndim, _parseScalarField(func))
    raise ValueError("Invalid function input parameter.")
 
def vectorField(idim, func=None, odim=None, address=None):
    if (func is None) == (address is None): raise ValueError("Must specify either func or address.")
    
    if address is not None:
        if odim is None: raise ValueError("Passing address requires also specifying odim.")
        return _vectorFieldFromAddress(idim, odim, address)
    if hasattr(func, "address"):
        if odim is None: raise ValueError("Passing address requires also specifying odim.")
        return _vectorFieldFromAddress(idim, odim, func.address)
    if isinstance(func, (list, tuple)) and reduce(lambda a, b: a and b, [isinstance(v, (bool, int, float)) for v in func]):
        return _vectorFieldFromTree(idim, _parseVectorField(str([float(f) for f in func])))
    if isinstance(func, str):
        return _vectorFieldFromTree(idim, _parseVectorField(func))
    raise ValueError("Invalid function input parameter.")
  
def implicitFunction(ndim, func=None, address=None):
    return implicitThreshold(scalarField(ndim, func, address), 0.5)
 
def domainIntegrand(ndim, callback, types, maxdiff, tmpdofs=0):
    return _domainIntegrandFromAddress(ndim, callback.address, types, maxdiff, tmpdofs) 
    
#del os, sys, path



import unittest
import sys

pattern = "*_test.py"

if len(sys.argv) > 1:
    pattern = sys.argv[1]

loader = unittest.TestLoader( )
suite = loader.discover('systemtests', pattern=pattern)
runner = unittest.TextTestRunner( )
result = runner.run(suite)
exit(not result.wasSuccessful())




# Move may of these into loadMkl?
import os
import sys
import ctypes
import glob
import re
import site
import mlhp

def loadMkl(*, mklRtPath=None, oneApiRoot=None, verbose=False):
    path = mklRtPath
    
    def printVerbose(message):
        if verbose:
            print("loadMkl: " + message)
    
    # Helper functions for oneAPI installation
    def findLibrary(folder):
        files = glob.glob(os.path.join(folder, "*mkl_rt.*.2"))
        printVerbose(f"Number of files matching: {len(files)}")
        for file in files:
            printVerbose(f"Matching file: {file}")
        return files[0] if len(files) else None
    
    def findOneApi(root):
        printVerbose(f"Looking in oneAPI root at {root}")
        return findLibrary(os.path.join(root, "mkl", "latest", "lib", "intel64"))
        
    # Helper function for python mkl package
    def findRecord(*args):
        directoryPath = os.path.join(*args, "mkl-*.dist-info")
        printVerbose(f"Searching at {directoryPath}")
        record = glob.glob(os.path.join(directoryPath, "RECORD"))
        if len(record) != 0:
            with open(record[0]) as file:
                relative = re.findall("(.*mkl_rt.*?),", file.read())
                filepath = glob.glob(os.path.join(directoryPath, "..", relative[0])) if len(relative) else []
                if len(filepath):
                    printVerbose(f"Found path: {filepath[0]}")
                    return filepath[0]
    
    # Try to locate mkl_rt
    if path is None and oneApiRoot is not None:
        printVerbose("Checking oneApiRoot function argument")
        path = findOneApi(oneApiRoot)
    
    if path is None and "ONEAPI_ROOT" in os.environ:
        path = findOneApi(os.environ["ONEAPI_ROOT"])
    
    if path is None:
        path = findRecord(sys.prefix, "[lL]ib", "site-packages")
    
    if path is None:
        path = findRecord(sys.prefix, "[lL]ib", "[pP]ython*", "site-packages")
    
    if path is None:
        path = findRecord(site.USER_SITE)
    
    # Load mkl library
    if path is None:
        raise ValueError("Unable to locate mkl_rt library. Did you pip install mkl?")
    else:
        return ctypes.cdll.LoadLibrary(path)
        printVerbose(f"Loading {path}")


class pardisoFactorize:
    def __init__(self, matrix, *, messageLevel=0, symmetric=None, mklLibrary=None):
        self.mkl = loadMkl( ) if mklLibrary is None else mklLibrary
        self.pt = (ctypes.c_int64 * 64)()
        self.matrix = matrix
        self.symmetric = symmetric
        self.messageLevel = ctypes.c_int64(messageLevel)
        self.iparm = (ctypes.c_int64 * 64)()
        self.iparm[ 0] = 1 # Do not use default values
        self.iparm[ 1] = 3 # Use parallel nested dissection with metis
        self.iparm[ 9] = 8 # Pivoting perturbation: 10^-8
        self.iparm[23] = 1 # Parallel factorization control: 1->improved two-level factorization algorithm
        self.iparm[26] = 0 # Do not perform matrix consistency check
        self.iparm[34] = 1 # Zero-based indexing
        self._callPardiso(12)

    def _callPardiso(self, pardisoPhase, rhs=None, solution=None):
        error = ctypes.c_int64(0)
        mattype = 11 if self.symmetric is None or not self.symmetric else 1
        mattype = ctypes.c_int64(2 if self.matrix.symmetricHalf else mattype)
        n = ctypes.c_int64(self.matrix.shape[0])
        a = ctypes.cast(self.matrix.data_address, ctypes.POINTER(ctypes.c_double))
        ia = ctypes.cast(self.matrix.indptr_address, ctypes.POINTER(ctypes.c_int64))
        ja = ctypes.cast(self.matrix.indices_address, ctypes.POINTER(ctypes.c_int64))
        b = ctypes.cast(0 if rhs is None else rhs, ctypes.POINTER(ctypes.c_double))
        x = ctypes.cast(0 if rhs is None else solution, ctypes.POINTER(ctypes.c_double))
        
        self.mkl.pardiso_64(
            self.pt,                                    # pt        Internal data structure
            ctypes.byref(ctypes.c_int64(1)),            # maxfact   Number of matrix factorizations
            ctypes.byref(ctypes.c_int64(1)),            # mnum      Matrix index
            ctypes.byref(mattype),                      # mtype     Matrix type
            ctypes.byref(ctypes.c_int64(pardisoPhase)), # phase     Pardiso solver phase
            ctypes.byref(n),                            # n         Number of rows
            a, ia, ja,                                  # a, ia, ja Nonzero values, inptr, and indices
            ctypes.byref(ctypes.c_int64(0)),            # perm      Permutation vector (not needed?)
            ctypes.byref(ctypes.c_int64(1)),            # nrhs      Number of right hand sides 
            self.iparm,                                 # iparm     Pardiso configuration parameters
            ctypes.byref(self.messageLevel),            # msglvl    Verbosity
            b, x,                                       # b, x      Right hand side(s) and Solution target(s)
            ctypes.byref(error)                         # errro     Error output
        )              
        
        if error.value != 0:
            errorMap = {
                -1: "input inconsistent", 
                -2: "not enough memory", 
                -3: "reordering problem", 
                -4: "Zero pivot, numerical factorization or iterative refinement problem. If the "
                    "error appears during the solution phase, try to change the pivoting perturba"
                    "tion (iparm[9]) and also increase the number of iterative refinement steps. "
                    "If it does not help, consider changing the scaling, matching and pivoting op"
                    "tions (iparm[10], iparm[12], iparm[20])", 
                -5: "unclassified (internal) error", 
                -6: "reordering failed", 
                -7: "diagonal matrix is singular", 
                -8: "32-bit integer overflow problem", 
                -9: "not enough memory for OOC", 
                -10: "error opening OOC files", 
                -11: "read/write error with OOC files", 
                -12: "pardiso_64 called from 32-bit library", 
                -13: "interrupted by the (user-defined) mkl_progress function", 
                -15: "internal error which can appear for iparm[23]=10 and iparm[12]=1. Try switc"
                     "h matching off (set iparm[12]=0 and rerun.)" 
            }
            
            errorMsg = errorMap[error.value] if error.value in errorMap else "unknown error code"
            
            raise RuntimeError(f"pardiso_64 error {error.value} ({errorMsg}) in phase {pardisoPhase}.")

    def cleanup(self):
        self._callPardiso(-1)
        
    def __call__(self, rhs, target=None):
        sol = mlhp.DoubleVector(self.matrix.shape[0]) if target is None else target
        vector_address = rhs.address if isinstance(rhs, mlhp.DoubleVector) else rhs
        target_address = sol.address if isinstance(sol, mlhp.DoubleVector) else sol
        
        self._callPardiso(33, vector_address, target_address)
        
        return sol

    def __del__(self):
        self.cleanup()


def pardisoSolve(matrix, vector, *, messageLevel=0, symmetric=None, mklLibrary=None):
    return pardisoFactorize(matrix, messageLevel=messageLevel, symmetric=symmetric, mklLibrary=mklLibrary)(vector)


def _checkSparseStatus(functionName, error):
    if error != 0:
        msgMap = {
            1: "SPARSE_STATUS_NOT_INITIALIZED (empty handle or matrix arrays)",
            2: "SPARSE_STATUS_ALLOC_FAILED (internal memory allocation failed)",
            3: "SPARSE_STATUS_INVALID_VALUE (input parameters contain an invalid value)",
            4: "SPARSE_STATUS_EXECUTION_FAILED (execution failed)",
            5: "SPARSE_STATUS_INTERNAL_ERROR (error in algorithm implementation occurred)",
            6: "SPARSE_STATUS_NOT_SUPPORTED (the requested operation is not supported)"
        }
        
        msg = msgMap[error] if error in msgMap else f" an unknown error code ({error})"
        
        raise RuntimeError(f"{functionName} failed with {msg}.")
    

class _mklCsrMatrix:
    def __init__(self, mkl, shape, indptr_address, indices_address, data_address, *, zeroBased=True, intType=ctypes.c_int64):
        self.ptr = ctypes.c_void_p()
        self.mkl = mkl
        a = ctypes.cast(data_address, ctypes.POINTER(ctypes.c_double))
        ia0 = ctypes.cast(indptr_address, ctypes.POINTER(intType))
        ia1 = ctypes.cast(indptr_address + ctypes.sizeof(intType), ctypes.POINTER(intType))
        ja = ctypes.cast(indices_address, ctypes.POINTER(intType))
        
        # https://www.intel.com/content/www/us/en/docs/onemkl/developer-reference-c/2025-1/mkl-sparse-create-csr.html
        error = mkl.mkl_sparse_d_create_csr(
            ctypes.POINTER(ctypes.c_void_p)(self.ptr),
            ctypes.c_int(0 if zeroBased else 1),
            intType(shape[0]), intType(shape[1]),
            ia0, ia1, ja, a
        )
        
        _checkSparseStatus("mkl_sparse_d_create_csr", error)
        
    def __del__(self):
        # https://www.intel.com/content/www/us/en/docs/onemkl/developer-reference-c/2025-1/mkl-sparse-destroy.html
        error = self.mkl.mkl_sparse_destroy(self.ptr)

        _checkSparseStatus("mkl_sparse_destroy", error)
        
class _matrixDescr(ctypes.Structure):
    _fields_ = [("type", ctypes.c_int), ("mode", ctypes.c_int), ("diag", ctypes.c_int)]
    
def _internalIlu(matrix, iluCallback, printInfo, returnMatrix, mklLibrary):
    
    mkl = loadMkl( ) if mklLibrary is None else mklLibrary
    
    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError("Non-square matrix in MKL ilu.")
        
    if matrix.symmetricHalf:
        raise ValueError("Symmetric matrix not implemented in MKL ilu.")
        
    # Create new 32-bit matrix data structure
    if matrix.nnz > 2**31:
        raise ValueError("To many entries to convert to 32-bit sparse data structure.")
    
    ia = (ctypes.c_int32 * (matrix.shape[0] + 1))()
    ja = (ctypes.c_int32 * matrix.nnz)()
    
    inptr = matrix.indptr_buffer
    indices = matrix.indices_buffer
    
    for i in range(len(ia)):
        ia[i] = inptr[i] + 1
        
    for i in range(len(ja)):
        ja[i] = indices[i] + 1
    
    # Prepare data call mkl function
    a = ctypes.cast(matrix.data_address, ctypes.POINTER(ctypes.c_double))
    ipar = (ctypes.c_int32 * 128)()
    dpar = (ctypes.c_double * 128)()

    ib, jb, bilu = iluCallback(mkl, matrix.shape[0], a, ia, ja, ipar, dpar)

    del ia, ja, ipar, dpar
    
    if printInfo:
        print("MKL Incomplete LU factorization:")
        print(f"    shape              : {tuple(matrix.shape)}")
        print(f"    number of nonzeros : {ib[-1]}")
        print(f"    average bandwith   : {ib[-1] / matrix.shape[0]:.5g}")
        print(f"    fill ratio         : {100 * ib[-1] / matrix.shape[0] / matrix.shape[1]:.4f} %")
        print(f"    heap memory usage  : {((len(ib) + len(jb)) * 4 + len(bilu) * 8) / 1e6:.5g} MB")
    
    if returnMatrix:
        LU = mlhp.allocateSparseMatrix(matrix.shape, ib[-1] - 1)
        
        LUindptr = LU.indptr_buffer
        LUindices = LU.indices_buffer
        LUdata = LU.data_buffer
    
        for i in range(len(ib)):
            LUindptr[i] = ib[i] - 1
            
        for i in range(ib[-1] - 1):
            LUindices[i] = jb[i] - 1
            LUdata[i] = bilu[i]
            
        return LU
    
    # Create linear operator with lower triangular and upper triangular substitution
    def iluSolve(vector_address, target_address, n):
        if n != iluSolve.shape[0] or n != iluSolve.shape[1]:
            raise ValueError("Inconsistent size in ilu substitution.")
        
        # See mkl_spblas.h
        SPARSE_OPERATION_NON_TRANSPOSE = 10
        SPARSE_MATRIX_TYPE_TRIANGULAR = 23
        SPARSE_FILL_MODE_LOWER = 40
        SPARSE_FILL_MODE_UPPER = 41
        SPARSE_DIAG_NON_UNIT = 50
        SPARSE_DIAG_UNIT = 51
        
        # Prepare triangular solve function arguments
        vector = ctypes.cast(vector_address, ctypes.POINTER(ctypes.c_double))
        tmp = ctypes.cast(iluSolve.tmp.address, ctypes.POINTER(ctypes.c_double))
        target = ctypes.cast(target_address, ctypes.POINTER(ctypes.c_double))
        
        # Lower triangular substitution with unit-diagonal, the SciPy equivalent of: 
        # tmp = scipy.linalg.solve_triangular(LU, vector, lower=True, unit_diagonal=True)
        # Where LU could be, e.g.: LU, P = scipy.linalg.lu_factor(matrix.todense())
        descr = _matrixDescr(type=SPARSE_MATRIX_TYPE_TRIANGULAR, mode=SPARSE_FILL_MODE_LOWER, diag=SPARSE_DIAG_UNIT)
        
        # https://www.intel.com/content/www/us/en/docs/onemkl/developer-reference-c/2025-1/mkl-sparse-trsv.html
        error = iluSolve.mkl.mkl_sparse_d_trsv(ctypes.c_int(SPARSE_OPERATION_NON_TRANSPOSE), 
            ctypes.c_double(1.0), iluSolve.A.ptr, descr, vector, tmp)
        
        _checkSparseStatus("mkl_sparse_d_trsv", error)
        
        # Upper triangular substitution, the SciPy equivalent of:
        # scipy.linalg.solve_triangular(LU, tmp, lower=False, unit_diagonal=False)
        descr = _matrixDescr(type=SPARSE_MATRIX_TYPE_TRIANGULAR, mode=SPARSE_FILL_MODE_UPPER, diag=SPARSE_DIAG_NON_UNIT)
        
        error = iluSolve.mkl.mkl_sparse_d_trsv(ctypes.c_int(SPARSE_OPERATION_NON_TRANSPOSE), 
            ctypes.c_double(1.0), iluSolve.A.ptr, descr, tmp, target)
        
        _checkSparseStatus("mkl_sparse_d_trsv", error)
    
    iluSolve.shape = matrix.shape
    iluSolve.mkl = mkl
    iluSolve.tmp = mlhp.DoubleVector(matrix.shape[0])
    iluSolve.data = (ib, jb, bilu)
    
    iluSolve.A = _mklCsrMatrix(
        mkl=mkl, shape=matrix.shape, 
        indptr_address=ctypes.cast(ib, ctypes.c_void_p).value, 
        indices_address=ctypes.cast(jb, ctypes.c_void_p).value, 
        data_address=ctypes.cast(bilu, ctypes.c_void_p).value, 
        zeroBased=False, intType=ctypes.c_int32
    )
    
    return mlhp.linearOperator_address(iluSolve)
    
def ilu0(matrix, *, printInfo=False, returnMatrix=False, mklLibrary=None):
    def ilu0Callback(mkl, n, a, ia, ja, ipar, dpar):

        bilu0 = (ctypes.c_double * (ia[n] - 1))()
        ierr = ctypes.c_int32(0)
        n = ctypes.c_int32(n)
        
        # https://www.intel.com/content/www/us/en/docs/onemkl/developer-reference-fortran/2025-1/dcsrilu0.html
        mkl.dcsrilu0(
            ctypes.byref(n),      # Number of rows
            a, ia, ja, bilu0,     # Nonzero source values, inptr, indices, and target values
            ipar, dpar,           # Integer and double precision set description
            ctypes.byref(ierr)    # Error output
        )    
        
        if ierr.value != 0:
            errorMap = {
                -101: "At least one diagonal element is omitted from the matrix in CSR3 format (see S"
                      "parse Matrix Storage Format).",
                -102: "Matrix contains a diagonal element with the value of zero.",
                -103: "Matrix contains a diagonal element which is so small that it could cause an ov"
                      "erflow, or that it would cause a bad approximation to ILU0.",
                -104: "Memory is insufficient for the internal work array.",
                -105: "Input matrix size n is less than or equal to 0.",
                -106: "Column indices ja are not in the ascending order."
            }

            errorMsg = errorMap[ierr.value] if ierr.value in errorMap else "unknown error code"
                
            raise RuntimeError(f"dcsrilu0 error {ierr.value} ({errorMsg}).")
            
        return ia, ja, bilu0
    return _internalIlu(matrix, ilu0Callback, printInfo, returnMatrix, mklLibrary)
    
def ilut(matrix, drop_tol=1e-4, fill_factor=10, *, printInfo=False, returnMatrix=False, mklLibrary=None):
    def ilutCallback(mkl, n, a, ia, ja, ipar, dpar):
        maxfil, tol = ilutCallback.maxfil, ilutCallback.tol        
        dpar[30] = tol
        ipar[30] = 1
        nnz = (2 * maxfil + 1) * n - maxfil * (maxfil + 1) + 1
        ib = (ctypes.c_int32 * (n + 1))()
        jb = (ctypes.c_int32 * nnz)()
        b = (ctypes.c_double * nnz)()
        ierr = ctypes.c_int32(0)
        n = ctypes.c_int32(n)
        tol = ctypes.c_double(tol)
        maxfil = ctypes.c_int(maxfil)
        
        # https://www.intel.com/content/www/us/en/docs/onemkl/developer-reference-fortran/2025-1/dcsrilut.html
        mkl.dcsrilut(
            ctypes.byref(n),      # Number of rows
            a, ia, ja,            # Nonzero source values, inptr, and indices
            b, ib, jb,            # Nonzero target for values, inptr, and indices
            ctypes.byref(tol),    # Tolerance above which to keep entries
            ctypes.byref(maxfil), # Half of the bandwidth (for each row: bandwidth <= 2 * maxfill + 1)
            ipar, dpar,           # Integer and double precision set description
            ctypes.byref(ierr)    # Error output
        )
        
        if ierr.value != 0:
            errorMap = {
                -101: "the number of elements in some matrix row specified in the sparse format is equ"
                      "al to or less than 0.",
                -102: "value of the computed diagonal element is less than the product of the given to"
                      "lerance and the current matrix row norm, and it cannot be replaced as ipar(31)="
                      "0.",
                -103: "element ia(i + 1) is less than or equal to the element ia(i) (see Sparse Matrix"
                      " Storage Format).",
                -104: "memory is insufficient for the internal work arrays.",
                -105: "input value of maxfil is less than 0.",
                -106: "size n of the input matrix is less than 0.",
                -107: "element of the array ja is less than 1, or greater than n (see Sparse Matrix St"
                      "orage Format).",
                 101: "value of maxfil is greater than or equal to n. The calculation is performed wit"
                      "h the value of maxfil set to (n-1).",
                 102: "value of tol is less than 0. The calculation is performed with the value of the"
                      " parameter set to (-tol)",
                 103: "absolute value of tol is greater than value of dpar(31); it can result in insta"
                      "bility of the calculation.",
                 104: "value of dpar(31) is equal to 0. It can cause calculations to fail."
            }

            errorMsg = errorMap[ierr.value] if ierr.value in errorMap else "unknown error code"
                
            raise RuntimeError(f"dcsrilut error {ierr.value} ({errorMsg}).")
            
        return ib, jb, b

    maxfil = int((fill_factor * (matrix.nnz - matrix.shape[0])) / (2 * matrix.shape[0])) + 1
    
    ilutCallback.maxfil = min(maxfil, matrix.shape[1] - 1)
    ilutCallback.tol = drop_tol
    
    return _internalIlu(matrix, ilutCallback, printInfo, returnMatrix, mklLibrary)

def linearOperator(matrix, *, symmetric=False, mklLibrary=None):
    mkl = loadMkl( ) if mklLibrary is None else mklLibrary
    
    def SpMV(vector_address, target_address, n):
        if n != SpMV.matrix.shape[0] or n != SpMV.matrix.shape[1]:
            raise ValueError("Inconsistent size in ilu substitution.")
        
        
        operation = ctypes.c_int(10) # transpose: 11
        alpha = ctypes.c_double(1.0)
        beta = ctypes.c_double(1.0)
        
        vector = ctypes.cast(vector_address, ctypes.POINTER(ctypes.c_double))
        target = ctypes.cast(target_address, ctypes.POINTER(ctypes.c_double))
        
        # https://www.intel.com/content/www/us/en/docs/onemkl/developer-reference-c/2025-1/mkl-sparse-mv.html
        error = SpMV.mkl.mkl_sparse_d_mv(operation, alpha, 
            SpMV.A.ptr, SpMV.descr, vector, beta, target)
        
        _checkSparseStatus("mkl_sparse_d_mv", error)
    
    # Copy to 32-bit data structure
    if matrix.nnz > 2**31:
        raise ValueError("To many entries to convert to 32-bit sparse data structure.")
    
    ia = (ctypes.c_int32 * (matrix.shape[0] + 1))()
    ja = (ctypes.c_int32 * matrix.nnz)()
    
    indptr, indices = matrix.indptr_buffer, matrix.indices_buffer
    
    for i in range(matrix.shape[0] + 1):
        ia[i] = indptr[i]
        
    for i in range(matrix.nnz):
        ja[i] = indices[i]
    
    # Prepare mkl data     
    SPARSE_MATRIX_TYPE_GENERAL = 20
    SPARSE_MATRIX_TYPE_SYMMETRIC = 21
    SPARSE_FILL_MODE_UPPER = 41
    SPARSE_FILL_MODE_FULL = 42
    SPARSE_DIAG_NON_UNIT = 50
    
    type = SPARSE_MATRIX_TYPE_SYMMETRIC if matrix.symmetricHalf or symmetric else SPARSE_MATRIX_TYPE_GENERAL
    mode = SPARSE_FILL_MODE_UPPER if matrix.symmetricHalf else SPARSE_FILL_MODE_FULL
        
    SpMV.mkl = mkl
    SpMV.matrix = matrix
    SpMV.data = (ia, ja)
    SpMV.descr = _matrixDescr(type=type, mode=mode, diag=SPARSE_DIAG_NON_UNIT)
    
    SpMV.A = _mklCsrMatrix(
        mkl=mkl, shape=matrix.shape, 
        indptr_address=ctypes.cast(ia, ctypes.c_void_p).value, 
        indices_address=ctypes.cast(ja, ctypes.c_void_p).value, 
        data_address=matrix.data_address, 
        zeroBased=True, intType=ctypes.c_int32
    )
    
    return mlhp.linearOperator_address(SpMV)
    

#import mlhp, numpy
#mesh = mlhp.makeRefinedGrid([4, 3, 5], [3, 2, 4])
#basis = mlhp.makeHpTrunkSpace(mesh, 2, nfields=mesh.ndim)
#dirichlet = mlhp.integrateDirichletDofs(mlhp.vectorField(mesh.ndim, [0.0]*mesh.ndim), basis, [0])
#matrix = mlhp.allocateSparseMatrix(basis, dirichlet[0])
#vector = mlhp.allocateRhsVector(matrix)
#kinematics = mlhp.smallStrainKinematics(mesh.ndim)
#material = mlhp.isotropicElasticMaterial(mlhp.scalarField(mesh.ndim, 1.0), mlhp.scalarField(mesh.ndim, 0.3))
#integrand = mlhp.staticDomainIntegrand(kinematics, material, mlhp.vectorField(mesh.ndim, [1.0] * mesh.ndim))
#mlhp.integrateOnDomain(basis, integrand, [matrix, vector], dirichletDofs=dirichlet)
#
#operator = linearOperator(matrix)
#print(operator(vector).tolist())

##mlhp.norm(mlhp.add(mkl.pardisoSolve(matrix, vector), mlhp.makeCGSolver(rtol=1e-14)(matrix, vector), -1.0))
#print(numpy.round(matrix.todense(), 2))
#
#import mklwrapper as mkl
#LU = mkl.ilut(matrix, returnMatrix=True, fill_factor=1e10, drop_tol=1e-200).todense()
#LU, perm = scipy.linalg.lu_factor(matrix.todense())
#tmp = scipy.linalg.solve_triangular(LU, vector, lower=True, unit_diagonal=True)
#result = scipy.linalg.solve_triangular(LU, tmp, lower=False, unit_diagonal=False)
#print(scipy.linalg.norm(mkl.pardisoSolve(matrix, vector) - result))
#LUop = mkl.ilut(matrix, returnMatrix=False, fill_factor=1e10, drop_tol=1e-200)
#print(scipy.linalg.norm(mkl.pardisoSolve(matrix, vector) - numpy.array(LUop(vector).tolist()))/scipy.linalg.norm(mkl.pardisoSolve(matrix, vector)))

if __name__ == "__main__":

    print("Testing mkl...", flush=True)
    
    shape = [6, 6]
    umatrix = mlhp.allocateSparseMatrix(shape, shape[0] * shape[1])
    smatrix = mlhp.allocateSparseMatrix(shape, (shape[0] * (shape[0] + 1)) // 2, symmetric=True)
    rhs = mlhp.allocateRhsVector(umatrix)

    uindex, sindex = 0, 0
    for i in range(shape[0]):
        rhs[i] = 0.7 * (shape[0] - i) * i / (shape[0]**2 / 4)  + 1.2 * i / shape[0] + 3.1
        
        umatrix.indptr_buffer[i] = uindex
        smatrix.indptr_buffer[i] = sindex
        
        for j in range(shape[1]):
            umatrix.indices_buffer[uindex] = j
            umatrix.data_buffer[uindex] = 2.3 / ( abs(i - j) + 1 )
        
            if j >= i:
                smatrix.indices_buffer[sindex] = j
                smatrix.data_buffer[sindex] = umatrix.data_buffer[uindex]
                
                sindex += 1
            uindex += 1
        
        umatrix.indptr_buffer[i + 1] = uindex
        smatrix.indptr_buffer[i + 1] = sindex

    expected = [0.42655929910747536, 0.4971564661821119 , 0.5773572916154234,
                0.6412062015071258 , 0.7093083642299071 , 1.068437405980519]

    computed1 = pardisoSolve(umatrix, rhs)
    computed2 = pardisoSolve(umatrix, rhs, symmetric=True)
    computed3 = pardisoSolve(smatrix, rhs)
    computed4 = pardisoSolve(smatrix, rhs, symmetric=True) # Shouldn't do anything
    computed5 = ilu0(umatrix)(rhs)
    computed6 = ilut(umatrix)(rhs)

    for computed, name in [(computed1, "Unsymmetric pardiso (general)"),
                           (computed2, "Unsymmetric pardiso (symmetric)"),
                           (computed3, "Symmetric half pardiso"),
                           (computed4, "Symmetric half pardiso (symmetric)"),
                           (computed5, "MKL ilu0"),
                           (computed6, "MKL ilut")]:
                           
        if sum((c - e)**2 for c, e in zip(computed, expected)) > 1e-12**2:
            raise ValueError(f"{name} solution incorrect.")
        else:
            print(f"{name} solution correct.", flush=True)

    
    solution = mlhp.DoubleVector(expected)
    
    mklSpuMV1 = linearOperator(umatrix, symmetric=False)
    computed1 = mklSpuMV1(solution)
    computed2 = linearOperator(umatrix, symmetric=True)(solution)
    computed3 = linearOperator(smatrix, symmetric=False)(solution)
    computed4 = linearOperator(smatrix, symmetric=True)(solution)

    for computed, name in [(computed1, "Unsymmetric SpMV (general)"),
                           (computed2, "Unsymmetric SpMV (symmetric)"),
                           (computed3, "Symmetric SpMV (general)"),
                           (computed4, "Symmetric SpMV (symmetric)")]:
                           
        if sum((c - e)**2 for c, e in zip(computed, rhs)) > 1e-12**2:
            raise ValueError(f"{name} incorrect.")
        else:
            print(f"{name} correct.", flush=True)



import mlhp
import math

D = 2

print("1. Preprocessing", flush=True)

nelements = 100
degree = 2
alpha = 1e-4
treedepth = degree + 3
filterElements = True

domain = mlhp.invert(mlhp.implicitSphere([0.0] * D, 0.5))

baseGrid = mlhp.makeGrid(ncells=[nelements] * D, lengths=[1.0] * D, origin=[0.0] * D)

if filterElements:
    baseGrid = mlhp.makeFilteredGrid(baseGrid, domain=domain, nseedpoints=degree + 2)

mesh = mlhp.makeRefinedGrid(baseGrid)
basis = mlhp.makeHpTrunkSpace(mesh, degrees=degree, nfields=D)

print(basis)

print("2. Assembly", flush=True)

# Determine Dirichlet dofs
leftDirichlet = mlhp.integrateDirichletDofs(mlhp.scalarField(D, 0.0), basis, [0], ifield=0)
bottomDirichlet = mlhp.integrateDirichletDofs(mlhp.scalarField(D, 0.0), basis, [2], ifield=1)

dirichletDofs=mlhp.combineDirichletDofs([leftDirichlet, bottomDirichlet])

# Allocate linear system for internal dofs
matrix = mlhp.allocateSparseMatrix(basis, dirichletDofs[0])
vector = mlhp.allocateRhsVector(matrix)

print(matrix)

# Assemble domain contribution
E = mlhp.scalarField(D, 206900 * 1e6)
nu = mlhp.scalarField(D, 0.29)
rhs = mlhp.vectorField(D, [0.0] * D)

kinematics = mlhp.smallStrainKinematics(D) 
constitutive = mlhp.planeStressMaterial(E, nu)
domainIntegrand = mlhp.staticDomainIntegrand(kinematics, constitutive, rhs)

#quadrature = mlhp.spaceTreeQuadrature(domain, depth=treedepth, epsilon=alpha)
quadrature = mlhp.momentFittingQuadrature(domain, depth=treedepth, epsilon=alpha) 

mlhp.integrateOnDomain(basis, domainIntegrand, [matrix, vector], quadrature=quadrature, dirichletDofs=dirichletDofs)

# Assemble traction integral
topQuadrature = mlhp.quadratureOnMeshFaces(mesh, [3], degree + 1)
forceIntegrand = mlhp.normalNeumannIntegrand(mlhp.scalarField(D, 100 * 1e6))
mlhp.integrateOnSurface(basis, forceIntegrand, [vector], topQuadrature, dirichletDofs=dirichletDofs)

print("3. Linear system solution", flush=True)
#import mklwrapper
#interiorDofs = mklwrapper.pardisoSolve(matrix, vector)

interiorDofs = mlhp.makeCGSolver(rtol=1e-12, maxiter=10000)(matrix, vector)
allDofs = mlhp.inflateDofs(interiorDofs, dirichletDofs)

print("4. Error integration", flush=True)

internalEnergy = mlhp.ScalarDouble(0.0)
energyIntegrand = mlhp.internalEnergyIntegrand(allDofs, kinematics, constitutive)
mlhp.integrateOnDomain(basis, energyIntegrand, [internalEnergy], quadrature=quadrature)

# This "exact" solution was computed with an overkill solution
relativeError = math.sqrt(abs(internalEnergy.get() - 5.78769965054534e4) / 5.78769965054534e4)

print(f"   Relative error: {100 * relativeError:.2e} percent")

# Check consistency with previous result
assert(abs(relativeError - 0.0019) < 1e-5)

print("5. File output", flush=True)

processors = [mlhp.solutionProcessor(D, allDofs, "Displacement"),
              mlhp.vonMisesProcessor(allDofs, kinematics, constitutive),
              mlhp.functionProcessor(domain)]

gridmesh = mlhp.gridOnCells([1 if degree == 1 else degree + 2] * D)
gridwriter = mlhp.PVtuOutput(filename="outputs/single")
mlhp.writeBasisOutput(basis, gridmesh, gridwriter, processors)



# This file is part of the mlhp project. License: See LICENSE

import mlhp
import unittest

def makeRefinement(position0, position1, time):
    sourcePosition = [(x1 - x0) * time + x0 for x0, x1 in zip(position0, position1)]
    
    refinement1Radii = [0.096, 0.048, 0.048]
    refinement2Radii = [0.052, 0.016, 0.016]

    center1 = [sourcePosition[0] - 0.6 * refinement1Radii[0]] + sourcePosition[1:]
    center2 = [sourcePosition[0] - 0.3 * refinement2Radii[0]] + sourcePosition[1:]
    
    domain1 = mlhp.implicitEllipsoid(center1, refinement1Radii)
    domain2 = mlhp.implicitEllipsoid(center2, refinement2Radii)
    
    refinement1 = mlhp.refineInsideDomain(domain1, 1)
    refinement2 = mlhp.refineInsideDomain(domain2, 2)
    
    return mlhp.refinementOr([refinement1, refinement2])
    
class LinearHeatTest (unittest.TestCase):
    def test_3D(self):
        
        D = 3
        theta = 0.5
        nsteps = 24
        
        lengths = [1.0, 0.4, 0.1]
        nelements = [10, 4, 1]
        degrees = [2, 2, 2]
        resolution = [2, 2, 2]
        
        duration = 1.0
        
        capacity = 1.0
        conductivity = 0.008
        sourceSigma = 0.02
        
        position0 = [0.2, lengths[1] / 2.0, lengths[2]]
        position1 = [0.8, lengths[1] / 2.0, lengths[2]]
        
        analytical = mlhp.makeAmLinearSolution(position0, position1, duration, 
            capacity, conductivity, sourceSigma, duration / 10.0, 0.0)
        
        initial = mlhp.scalarField(D, 0.0)
        
        dirichletSide = [i for i in range(2 * D - 1)]
        
        # Time integration
        grid0 = mlhp.makeRefinedGrid(nelements, lengths)
        grid0.refine(makeRefinement(position0, position1, 0.0))
        
        basis0 = mlhp.makeHpTrunkSpace(grid0, mlhp.LinearGrading(degrees))

        dofs0 = mlhp.projectOnto(basis0, initial)
        
        processors = [mlhp.solutionProcessor(D, dofs0, "Temperature"),
                      mlhp.functionProcessor(mlhp.sliceLast(analytical.solution, 0.0), "Analytical"),
                      mlhp.functionProcessor(mlhp.sliceLast(analytical.source, 0.0), "Source")]
                           
        postmesh = mlhp.gridOnCells(resolution)
        writer = mlhp.VtuOutput(f'outputs/linear_heat_{0}')
        
        mlhp.writeBasisOutput(basis0, postmesh, writer, processors)        
        
        dt = duration / nsteps
        ndof = 0
        integrals = [0.0, 0.0, 0.0]
        
        for istep in range(nsteps):
            time0, time1 = istep * dt, (istep + 1) * dt
            
            grid1 = mlhp.makeRefinedGrid(nelements, lengths)
            grid1.refine(makeRefinement(position0, position1, time1))
            
            basis1 = mlhp.makeHpTrunkSpace(grid1, mlhp.LinearGrading(degrees))

            print("Time step " + str(istep + 1) + " / " + str(nsteps) +
                  " (" + str(basis1.ndof()) + " number of unknowns)")
               
            dirichlet = mlhp.integrateDirichletDofs(mlhp.sliceLast(
                analytical.solution, time1), basis1, dirichletSide)

            matrix = mlhp.allocateSparseMatrix(basis1, dirichlet[0])
            vector = mlhp.allocateRhsVector(matrix)
            
            integrand = mlhp.transientPoissonIntegrand(mlhp.scalarField(D + 1, capacity),
                mlhp.scalarField(D + 1, conductivity), analytical.source, dofs0, [time0, time1], theta)
                         
            mlhp.integrateOnDomain(basis0, basis1, integrand, [matrix, vector], dirichletDofs=dirichlet)
            
            dofs1 = mlhp.inflateDofs(mlhp.makeCGSolver()(matrix, vector), dirichlet)
            
            # Error integration (excluding initial condition         
            l2ErrorIntegrand = mlhp.l2ErrorIntegrand(dofs1, mlhp.sliceLast(analytical.solution, time1))
            l2Integrals = mlhp.makeScalars(3)
            
            mlhp.integrateOnDomain(basis1, l2ErrorIntegrand, l2Integrals)
            
            factor = dt if istep + 1 < nsteps else dt / 2.0
            
            integrals = [E + factor * Ec.get() for E, Ec in zip(integrals, l2Integrals)]
                     
            # Vtu postprocessing
            processors = [mlhp.solutionProcessor(D, dofs1, "Temperature"),
                          mlhp.functionProcessor(mlhp.sliceLast(analytical.solution, time1), "Analytical"),
                          mlhp.functionProcessor(mlhp.sliceLast(analytical.source, time1), "Source")]
            
            postmesh = mlhp.gridOnCells(resolution)
            writer = mlhp.VtuOutput(f'outputs/linear_heat_{istep + 1}')
                
            mlhp.writeBasisOutput(basis1, postmesh, writer, processors)             
            
            ndof += basis1.ndof()
            dofs0 = dofs1
            basis0 = basis1
        
        self.assertEqual(ndof, 33436)
        self.assertAlmostEqual(integrals[0], 2.7342053949**2, delta=1e-8)
        self.assertAlmostEqual(integrals[1], 2.7532175961**2, delta=1e-8)
        self.assertAlmostEqual(integrals[2], 0.0644606778**2, delta=1e-7)
        self.assertAlmostEqual(integrals[2] / integrals[1], 0.0234128526**2, delta=1e-7)
        



from __future__ import annotations

import pytest

from pybind11_tests import class_sh_trampoline_basic as m


class PyDrvd0(m.Abase0):
    def __init__(self, val):
        super().__init__(val)

    def Add(self, other_val):
        return self.Get() * 100 + other_val


class PyDrvd1(m.Abase1):
    def __init__(self, val):
        super().__init__(val)

    def Add(self, other_val):
        return self.Get() * 200 + other_val


def test_drvd0_add():
    drvd = PyDrvd0(74)
    assert drvd.Add(38) == (74 * 10 + 3) * 100 + 38


def test_drvd0_add_in_cpp_raw_ptr():
    drvd = PyDrvd0(52)
    assert m.AddInCppRawPtr(drvd, 27) == ((52 * 10 + 3) * 100 + 27) * 10 + 7


def test_drvd0_add_in_cpp_shared_ptr():
    while True:
        drvd = PyDrvd0(36)
        assert m.AddInCppSharedPtr(drvd, 56) == ((36 * 10 + 3) * 100 + 56) * 100 + 11
        return  # Comment out for manual leak checking (use `top` command).


def test_drvd0_add_in_cpp_unique_ptr():
    while True:
        drvd = PyDrvd0(0)
        with pytest.raises(ValueError) as exc_info:
            m.AddInCppUniquePtr(drvd, 0)
        assert (
            str(exc_info.value)
            == "Alias class (also known as trampoline) does not inherit from"
            " py::trampoline_self_life_support, therefore the ownership of this"
            " instance cannot safely be transferred to C++."
        )
        return  # Comment out for manual leak checking (use `top` command).


def test_drvd1_add_in_cpp_unique_ptr():
    while True:
        drvd = PyDrvd1(25)
        assert m.AddInCppUniquePtr(drvd, 83) == ((25 * 10 + 3) * 200 + 83) * 100 + 13
        return  # Comment out for manual leak checking (use `top` command).



import mlhp 
import sys 
import numpy 
import gmshparser

assert(len(sys.argv) == 2)

print("1. Reading mesh", flush=True)

mesh = gmshparser.parse(sys.argv[1])

vertices = numpy.array([node.get_coordinates() for entity in mesh.get_node_entities() for node in entity.get_nodes()])
cells = [element.get_connectivity() for entity in mesh.get_element_entities() for element in entity.get_elements()]
types = numpy.array([entity.get_element_type() for entity in mesh.get_element_entities() for element in entity.get_elements()])

# Mesh is 3D
if numpy.any(types == 4) or numpy.any(types == 5):
    vertices = [vertex[:3] for vertex in vertices]
    # https://gmsh.info/dev/doc/texinfo/gmsh.pdf page 362/372
    hexes = [[cell[0], cell[1], cell[3], cell[2], cell[4], cell[5], cell[7], cell[6]]\
        for type, cell in zip(types, cells) if type == 5]
    cells = hexes + [cell for type, cell in zip(types, cells) if type == 4]
    
# Mesh is 2D
elif numpy.any(types == 2) or numpy.any(types == 3):
    vertices = [vertex[:2] for vertex in vertices]
    quads = [[cell[0], cell[1], cell[3], cell[2]] for type, cell in zip(types, cells) if type == 3]
    cells = quads + [cell for type, cell in zip(types, cells) if type == 2]
else:
    RuntimeError("Did not find 2- or 3D cubic or simplex elements")
    
offsets = numpy.cumsum([0] + [len(cell) for cell in cells])
cells = numpy.array([id for cell in cells for id in cell]) - 1

print("2. Setting up mesh and basis", flush=True)

mesh = mlhp.makeUnstructuredMesh(vertices, cells, offsets)

ndim = mesh.ndim
nfields = ndim

basis = mlhp.makeUnstructuredBasis(mesh, nfields)

print(mesh)
print(basis)

print("3. Computing dirichlet boundary conditions", flush=True)

dirichlet = mlhp.integrateDirichletDofs(mlhp.vectorField(ndim, [0.0] * nfields), basis, [0])

print("4. Setting up physics", flush=True)

E = mlhp.scalarField(ndim, 200 * 1e9)
nu = mlhp.scalarField(ndim, 0.3)
rhs = mlhp.vectorField(ndim, [0.0] * (ndim - 1) + [-1.0])

if ndim == 3:
    constitutive = mlhp.isotropicElasticMaterial(E, nu)
if ndim == 2:
    constitutive = mlhp.planeStressMaterial(E, nu)

kinematics = mlhp.smallStrainKinematics(ndim) 
integrand = mlhp.staticDomainIntegrand(kinematics, constitutive, rhs)

print("5. Allocating linear system", flush=True)

matrix = mlhp.allocateSparseMatrix(basis, dirichlet[0])
vector = mlhp.allocateRhsVector(matrix)

print("6. Integrating linear system", flush=True)

mlhp.integrateOnDomain(basis, integrand, [matrix, vector], 
    dirichletDofs=dirichlet)

print("7. Solving linear system", flush=True)

P = mlhp.diagonalPreconditioner(matrix)

interiorDofs, norms = mlhp.cg(matrix, vector, M=P, maxiter=5000, residualNorms=True)

allDofs = mlhp.inflateDofs(interiorDofs, dirichlet)

print("8. Postprocessing solution", flush=True)

processors = [mlhp.solutionProcessor(ndim, allDofs, "Displacement"),
              mlhp.vonMisesProcessor(allDofs, kinematics, constitutive)]

postmesh = mlhp.gridOnCells([1] * ndim)
writer = mlhp.PVtuOutput(filename="outputs/example_gmsh")
            
mlhp.writeBasisOutput(basis, postmesh, writer, processors)



import mlhp, math

D = 3

print("1. Setting up mesh and basis", flush=True)

degree = 3
treedepth = 1
alpha = 1e-4

# Parse implicit function string (alternatively: use numba, see end of file)
Lx, Ly, Lz, t, pi = 1, 1, 1, 0.2, math.pi
term1 = f"cos(2*{pi}*x/{Lx}) * sin(2*{pi}*y/{Ly})"
term2 = f"cos(2*{pi}*y/{Ly}) * sin(2*{pi}*z/{Lz})"
term3 = f"cos(2*{pi}*z/{Lz}) * sin(2*{pi}*x/{Lx})"

domain = mlhp.implicitFunction(3, f"abs({term1} + {term2} + {term3}) < {t}")

grid = mlhp.makeGrid(ncells=[20] * D, lengths=[1.9] * D, origin=[-0.45] * D)
print(grid)

grid = mlhp.makeRefinedGrid(mlhp.makeFilteredGrid(grid, domain=domain, nseedpoints=degree + 2))
basis = mlhp.makeHpTrunkSpace(grid, degrees=degree, nfields=D)

print(grid)
print(basis)

print("2. Computing dirichlet boundary conditions", flush=True)

left = mlhp.integrateDirichletDofs(mlhp.vectorField(D, [0.0] * D), basis, [0])
right = mlhp.integrateDirichletDofs(mlhp.scalarField(D, 1e-3), basis, [1], ifield=0)

dirichlet=mlhp.combineDirichletDofs([left, right])

print("3. Setting up physics", flush=True)

E = mlhp.scalarField(D, 200 * 1e9)
nu = mlhp.scalarField(D, 0.3)
rhs = mlhp.vectorField(D, [0.0, 0.0, 0])

kinematics = mlhp.smallStrainKinematics(D) 
constitutive = mlhp.isotropicElasticMaterial(E, nu)
integrand = mlhp.staticDomainIntegrand(kinematics, constitutive, rhs)

print("4. Allocating linear system", flush=True)

matrix = mlhp.allocateSparseMatrix(basis, dirichlet[0])
vector = mlhp.allocateRhsVector(matrix)

print(matrix)

print("5. Integrating linear system", flush=True)

#quadrature = mlhp.spaceTreeQuadrature(domain, depth=treedepth, epsilon=alpha)
quadrature = mlhp.momentFittingQuadrature(domain, depth=treedepth, epsilon=alpha)

# To extract quadrature points with integration order 3 (plot uses matplotlib and assumes D = 2):
# partitions = [quadrature.evaluate(mesh, icell, [3, 3]) for icell in range(mesh.ncells())]
# coordinates = [xyz for cell in partitions for partition in cell for xyz in partition.xyz]
# plt.plot([c[0] for c in coordinates], [c[1] for c in coordinates], 'x')

mlhp.integrateOnDomain(basis, integrand, [matrix, vector], 
    dirichletDofs=dirichlet, quadrature=quadrature)

print("6. Solving linear system", flush=True)

P = mlhp.additiveSchwarzPreconditioner(matrix, basis, dirichlet[0])
#P = mlhp.diagonalPreconditioner(matrix)

interiorDofs, norms = mlhp.cg(matrix, vector, M=P, maxiter=10000, residualNorms=True)

del matrix, P

allDofs = mlhp.inflateDofs(interiorDofs, dirichlet)

print("7. Postprocessing solution", flush=True)

processors = [mlhp.solutionProcessor(D, allDofs, "Displacement"),
              mlhp.vonMisesProcessor(allDofs, kinematics, constitutive),
              mlhp.functionProcessor(domain)]

gridmesh = mlhp.gridOnCells(mlhp.degreeOffsetResolution(basis), mlhp.PostprocessTopologies.Volumes)
surfmesh = mlhp.marchingCubesBoundary(domain, [degree + 2]*D)

gridwriter = mlhp.PVtuOutput(filename="outputs/gyroid_mesh")
surfwriter = mlhp.PVtuOutput(filename="outputs/gyroid_surf")
            
mlhp.writeBasisOutput(basis, gridmesh, gridwriter, processors)
mlhp.writeBasisOutput(basis, surfmesh, surfwriter, processors)

# # Alternatively: use numba to pass compiled implicit function
# from numba import cfunc, types, carray
# 
# Lx, Ly, Lz, t = 1, 1, 1, 0.2
# 
# @cfunc(types.float64(types.CPointer(types.double), types.int64))
# def gyroid(xyz, ndim): 
#     x, y, z = carray(xyz, ndim)[0], carray(xyz, ndim)[1], carray(xyz, ndim)[2]
#     
#     term1 = math.cos(2 * math.pi * x / Lx) * math.sin(2 * math.pi * y / Ly)
#     term2 = math.cos(2 * math.pi * y / Ly) * math.sin(2 * math.pi * z / Lz)
#     term3 = math.cos(2 * math.pi * z / Lz) * math.sin(2 * math.pi * x / Lx)
#     
#     return abs(term1 + term2 + term3) < t
# 
# domain = mlhp.implicitFunction(3, gyroid)



# This file is part of the mlhp project. License: See LICENSE

import mlhp
import unittest

def compute(D, refinementDepth):

    analytical = mlhp.makeSingularSolution(D).solution

    # Domain and discretization
    nelements = [4] * D
    lengths = [1.0] * D

    sphere = mlhp.implicitSphere([0.0] * D, 0.0)
    strategy = mlhp.refineTowardsBoundary(sphere, refinementDepth)

    grid = mlhp.makeRefinedGrid(nelements, lengths)
    grid.refine(strategy)

    basis = mlhp.makeHpTrunkSpace(grid, mlhp.LinearGrading(1))

    # Compute indices and values for dirichlet boundary condition
    dofs = mlhp.projectOnto(basis, analytical)
    
    # Vtk Postprocessing
    processors = [mlhp.solutionProcessor(D, dofs),
                  mlhp.functionProcessor(analytical, "Analytical")]

    postmesh = mlhp.gridOnCells([refinementDepth + 1] * D)
    writer = mlhp.VtuOutput(f'outputs/singular_L2_{D}D_{refinementDepth}')
            
    mlhp.writeBasisOutput(basis, postmesh, writer, processors)
        
    # Error computation
    l2ErrorIntegrand = mlhp.l2ErrorIntegrand(dofs, analytical)
    l2Integrals = mlhp.makeScalars(3)
    
    mlhp.integrateOnDomain(basis, l2ErrorIntegrand, l2Integrals)
                               
    # Return number of dofs and || u_fe - u_ex ||^2 
    return [ len(dofs), l2Integrals[2].get() ]
    
class SingularL2Test (unittest.TestCase):
    
    def computeStudy(self, D, expected, tol):
        expectedNDofs, expectedL2 = expected
        
        for iStudy, expectedNDof in enumerate(expectedNDofs):
            computedNDof, computedL2 = compute(D, iStudy)
            
            self.assertEqual(computedNDof, expectedNDof)
            self.assertAlmostEqual(computedL2, expectedL2[iStudy], delta=tol)
        
    def test_1D(self):
    
        expectedNDof = [5, 9, 14, 20, 27]
        l2 = [3.08705e-06, 6.21486e-07, 1.25989e-07, 2.55838e-08, 5.19513e-09]

        self.computeStudy(1, [expectedNDof, l2], 1e-10)
             
    def test_2D(self):
    
        expectedNDof = [25, 66, 113, 181, 273]
        l2 = [1.42219e-06, 1.2541e-06, 1.67571e-07, 2.11436e-08, 2.68362e-09]

        self.computeStudy(2, [expectedNDof, l2], 1e-10)
             
    def test_3D(self):
    
        expectedNDof = [125, 429, 754, 1337]
        l2 = [7.35596e-07, 2.28911e-07, 1.66542e-08, 1.27925e-09]

        self.computeStudy(3, [expectedNDof, l2], 1e-10)
     



# This file is part of the mlhp project. License: See LICENSE

import mlhp, math
import unittest

class PressurizedSphereTest(unittest.TestCase):
    def test_elastic(self):
    
        D = 3
        
        # inner, outer, and plastic radii
        a, b, c = 0.1, 0.2, 0.1 
        
        youngsModulus = 200 * 1e9
        sigmaY = 200 * 1e6
        P = 2 * sigmaY * math.log(c / a) + 2 * sigmaY / 3 * ( 1 - c**3 / b**3 )
        
        # Domain setup
        inner = mlhp.implicitSphere( [0.0, 0.0, 0.0], a )
        outer = mlhp.implicitSphere( [0.0, 0.0, 0.0], b )
        
        domain = mlhp.implicitSubtraction([outer, inner])
        
        origin, max = [0.0] * D, [1.2 * b] * D
        lengths = [m - o for o, m in zip(origin, max)]
        
        # Discrtization setup
        nelements = [20] * D
        polynomialDegree = 1
        alphaFCM = 1e-5
        
        grid = mlhp.makeGrid(nelements, lengths, origin)
        grid = mlhp.makeFilteredGrid(grid, domain=domain, nseedpoints=polynomialDegree + 2)
        grid = mlhp.makeRefinedGrid(grid)
        
        basis = mlhp.makeHpTrunkSpace(grid, degrees=polynomialDegree, nfields=D)

        # Constrain normal displacement on three sides
        face0 = mlhp.integrateDirichletDofs(mlhp.scalarField(D, 0.0), basis, [0 * 2], ifield=0)
        face1 = mlhp.integrateDirichletDofs(mlhp.scalarField(D, 0.0), basis, [1 * 2], ifield=1)
        face2 = mlhp.integrateDirichletDofs(mlhp.scalarField(D, 0.0), basis, [2 * 2], ifield=2)
    
        dirichlet = mlhp.combineDirichletDofs([face0, face1, face2])

        # Allocate sparse linear system
        matrix = mlhp.allocateSparseMatrix(basis, dirichlet[0])
        vector = mlhp.allocateRhsVector(matrix)

        # Recover inner boundary and impose pressure
        resolution = [polynomialDegree + 3] * D
        triangulation, celldata = mlhp.marchingCubesBoundary(grid, inner, resolution)

        quadrature = mlhp.triangulationQuadrature(triangulation, celldata, polynomialDegree + 1)
        integrand = mlhp.normalNeumannIntegrand(mlhp.scalarField(D, -P))

        mlhp.integrateOnSurface(basis, integrand, [vector], quadrature, dirichletDofs=dirichlet)
        
        # Assemble domain contributions
        E = mlhp.scalarField(D, youngsModulus)
        nu = mlhp.scalarField(D, 0.3)
        rhs = mlhp.vectorField(D, [0.0, 0.0, 0.0])

        kinematics = mlhp.smallStrainKinematics(D) 
        constitutive = mlhp.isotropicElasticMaterial(E, nu)
        integrand = mlhp.staticDomainIntegrand(kinematics, constitutive, rhs)

        quadrature = mlhp.spaceTreeQuadrature(domain, depth=polynomialDegree + 1, epsilon=alphaFCM)

        mlhp.integrateOnDomain(basis, integrand, [matrix, vector], quadrature=quadrature, dirichletDofs=dirichlet)

        # Solve linear system
        P = mlhp.diagonalPreconditioner(matrix)

        interiorDofs = mlhp.cg(matrix, vector, rtol=1e-12, M=P, maxiter=2000)

        dofs = mlhp.inflateDofs(interiorDofs, dirichlet)
        
        # Postprocess solution
        #gradient = mlhp.projectGradient(basis, dofs, quadrature)

        processors = [mlhp.solutionProcessor(D, dofs, "Displacement"),
        #              mlhp.vonMisesProcessor(gradient, kinematics, constitutive, "ProjVonMises"),
                      mlhp.vonMisesProcessor(dofs, kinematics, constitutive)]
        surfmesh = mlhp.marchingCubesBoundary(domain, [5]*D)
        writer = mlhp.PVtuOutput(filename="outputs/pressurized_sphere_surface")

        mlhp.writeBasisOutput(basis, surfmesh, writer, processors)
        
        self.assertEqual(len(dofs), 9876)
        self.assertAlmostEqual(mlhp.norm(dofs), 0.0015142966728928131, delta=1e-7)
        



# This file is part of the mlhp project. License: See LICENSE

import mlhp
import unittest
import math

class CollectionTest(unittest.TestCase):
    def test_integrateScalarFunction2D(self):
        D = 2
        radius = 0.8
        order = 2

        domain = mlhp.implicitSphere([0.0] * D, radius)
        mesh = mlhp.makeGrid([axis + 3 for axis in range(D)], [2] * D, [-1] * D)
        basis = mlhp.makeDummyBasis(mesh)
        target = mlhp.ScalarDouble()

        quadratureOrder = mlhp.relativeQuadratureOrder(D, order)
        quadratureScheme = mlhp.momentFittingQuadrature(domain, depth=5, epsilon=0.0)
        #quadratureScheme = mlhp.spaceTreeQuadrature(domain, depth=5, epsilon=0.0)
        
        integrand = mlhp.functionIntegrand(mlhp.scalarField(D, 1.0))
        
        mlhp.integrateOnDomain(basis, integrand, [target], quadrature=quadratureScheme, orderDeterminor=quadratureOrder)

        expected = math.pi**(D / 2) / math.gamma(D / 2 + 1) * radius**D
        error = abs(expected - target.get()) / expected
        
        #print(error)
        
        self.assertAlmostEqual(error, 0.0007483838265199594, places=6)
        
    def test_integrateVectorFunction3D(self):
        D = 3
        radius = 0.83
        order = 2

        domain = mlhp.implicitSphere([0.0] * D, radius)
        
        r = "((x**2 + y**2 + z**2)**(1/2))"
        
        f0 = f"-2 * {r}**3 + 3 * {r}**2 + 1"
        f1 = f"4 * (1 - {r}**2)"
        
        function = mlhp.vectorField(D, f"[{f0}, {f1}]")
        
        mesh = mlhp.makeGrid([6 - axis for axis in range(D)], [2] * D, [-1] * D)
        basis = mlhp.makeDummyBasis(mesh)
        targets = [mlhp.ScalarDouble() for _ in range(2)]

        #meshcreator = mlhp.marchingCubesVolume(function=domain, resolution=[4] * D, coarsen=True, meshBothSides=False)
        #quadratureScheme = mlhp.cellMeshQuadrature(meshcreator)
        #quadratureScheme = mlhp.momentFittingQuadrature(domain, depth=3, epsilon=0.0)
        quadratureScheme = mlhp.spaceTreeQuadrature(domain, depth=4, epsilon=0.0)
        
        quadratureOrder = mlhp.relativeQuadratureOrder(D, order)
        integrand = mlhp.functionIntegrand(function)
        
        mlhp.integrateOnDomain(basis, integrand, targets, quadrature=quadratureScheme, orderDeterminor=quadratureOrder)
        
        result = [target.get() for target in targets]
        
        expected0 = 4 * math.pi * (-2/6 * radius**6 + 3/5 * radius**5 + 1/3 * radius**3)
        expected1 = 16 * math.pi * (radius**3 / 3 - radius**5 / 5)
        
        error0 = abs(expected0 - targets[0].get()) / expected0
        error1 = abs(expected1 - targets[1].get()) / expected1
        
        #print(error0, error1)
        
        self.assertAlmostEqual(error0, 3.404156148046146e-05, places=6)
        self.assertAlmostEqual(error1, 5.345606556139261e-06, places=6)
        
        # Integral in spherical coordinates:
        # int_rho(0, 2*pi) int_phi(0, pi) int_r(0, R)  f * r**2 * sin(phi) * dr * dphi * drho
        #
        # f = f1: int_rho(0, 2*pi) int_phi(0, pi) int_r(0, R) (-2 * r**3 + 3 * r**2 + 1) * r**2 * sin(phi) * dr * dphi * drho =
        #         int_rho(0, 2*pi) int_phi(0, pi) (-2/6 * R**6 + 3/5 * R**5 + 1/3 * R**3) * [-cos(phi)](0, pi) * drho =
        #         int_rho(0, 2*pi) 2 * (-2/6 * R**6 + 3/5 * R**5 + 1/3 * R**3) * drho =
        #         4*pi * (-2/6 * R**6 + 3/5 * R**5 + 1/3 * R**3)
        #      
        # f = f2: int_rho(0, 2*pi) int_phi(0, pi) int_r(0, R) 4 * (1 - r**2) * r**2 * sin(phi) * dr * dphi * drho =
        #         int_rho(0, 2*pi) int_phi(0, pi) 4/3 * R**3 - 4/5 * R**5 * sin(phi) * dphi * drho =
        #         16 * pi * (R**3 / 3 - R**5 / 5)
        
        



# This file is part of the mlhp project. License: See LICENSE

import mlhp
import operator
import unittest

from functools import reduce 

def computeImmersedBeam():

    # Prepare cube vertices [(0, 0, 0), (0, 0, 1), (0, 1, 0), (0, 1, 1), (1, 0, 0), ..., (1, 1, 1)]
    vertices = [(float(i), float(j), float(k)) for i in range(2) for j in range(2) for k in range(2)]

    faces = [mlhp.Triangulation(vertices, [[0, 2, 1], [2, 3, 1]]), # left
             mlhp.Triangulation(vertices, [[4, 6, 5], [6, 7, 5]]), # right
             mlhp.Triangulation(vertices, [[0, 1, 4], [4, 1, 5]]), # front
             mlhp.Triangulation(vertices, [[2, 3, 6], [6, 3, 7]]), # back
             mlhp.Triangulation(vertices, [[0, 4, 2], [4, 6, 2]]), # bottom
             mlhp.Triangulation(vertices, [[1, 5, 3], [5, 7, 3]])] # top

    # Transform computational domain to not be axis-aligned
    scale = mlhp.scaling([0.8, 0.9, 1.3])
    rotate = mlhp.rotation([3.0, 1.0, 2.0], 2.9)
    translate = mlhp.translation([-2.0, 3.2, 0.4])
    transformation = mlhp.concatenate([scale, rotate, translate])

    for i, face in enumerate(faces):
        face.transform(transformation)
        #face.writeVtu(f"outputs/beam_face_{i}.vtu")
        
    # Create mesh and finite element basis
    bounds = [t.boundingBox() for t in faces]

    bounds = [[min([b[0][axis] for b in bounds]) for axis in range(3)],
              [max([b[1][axis] for b in bounds]) for axis in range(3)]]

    lengths = [(1 + 2e-6) * (b - a) for a, b in zip(*bounds)]
    origin = [a - 1e-6 * (b - a) for a, b in zip(*bounds)]

    targetNElements = 100
    elementSize = (reduce(operator.mul, lengths) / targetNElements)**(1 / 3)
    nelements = [int(l / elementSize) + 1 for l in lengths]
    
    degree = 3
    epsilon = 1e-6
    
    mesh = mlhp.makeRefinedGrid(nelements, lengths, origin)
    basis = mlhp.makeHpTrunkSpace(mesh, degrees=degree, nfields=3)

    #print(mesh)
    #print(basis)

    # Linear system allocation
    matrix = mlhp.allocateSparseMatrix(basis, symmetric=False)######################################3
    vector = mlhp.allocateRhsVector(matrix)

    #print(matrix, flush=True)

    # Domain quadrature
    E = mlhp.scalarField(3, 1.0)
    nu = mlhp.scalarField(3, 0.3)

    domain = mlhp.implicitTransformation(mlhp.implicitCube([0.0] * 3, [1.0] * 3), transformation)

    kinematics = mlhp.smallStrainKinematics(ndim=3)
    material = mlhp.isotropicElasticMaterial(E, nu)
    domainIntegrand = mlhp.staticDomainIntegrand(kinematics, material)
    
    #domainQuadrature = mlhp.spaceTreeQuadrature(domain, depth=degree + 2, epsilon=epsilon)
    domainQuadrature = mlhp.momentFittingQuadrature(domain, depth=3, epsilon=epsilon)

    mlhp.integrateOnDomain(basis, domainIntegrand, [matrix, vector], domainQuadrature)

    # Boundary quadrature
    surfaceQuadratures = [mlhp.triangulationQuadrature(*mlhp.intersectTriangulationWithMesh(mesh, face), degree + 2) for face in faces]

    tractionFace, reactionFace = 5, 4

    # Fix all other faces
    for iface in [i for i in range(6) if i != tractionFace]:
        boundaryIntegrand0 = mlhp.l2NormalIntegrand(lhs=mlhp.scalarField(3, 1e6))
        mlhp.integrateOnSurface(basis, boundaryIntegrand0, [matrix], surfaceQuadratures[iface])
        
    # Impose traction right
    traction = faces[tractionFace].integrateNormalComponents()
    tractionNormSquared = reduce(operator.add, [x**2 for x in traction])
    traction = [x / tractionNormSquared for x in traction]

    boundaryIntegrand1 = mlhp.neumannIntegrand(mlhp.vectorField(3, traction))
    #boundaryIntegrand1 = mlhp.l2NormalIntegrand(rhs=mlhp.scalarField(3, tractionNormSquared**(1/2)))
    
    mlhp.integrateOnSurface(basis, boundaryIntegrand1, [vector], surfaceQuadratures[tractionFace])

    # Linear solution
    
    #import mklwrapper as mkl
    #solution = mkl.pardisoSolve(matrix, vector, symmetric=True)
    
    preconditioner = mlhp.additiveSchwarzPreconditioner(matrix, basis)
    
    solution, norms = mlhp.cg(matrix, vector, M=mlhp.linearOperator(preconditioner), residualNorms=True)
    
    #print(f"{len(norms) = }")

    # Reaction force computation method 1 (add entries of residual vector)
    one = mlhp.projectOnto(basis, mlhp.vectorField(3, [1] * basis.nfields()))
    res = mlhp.allocateRhsVector(matrix)

    support = []

    for ifield in range(3):
        supportIntegrand0 = mlhp.neumannIntegrand(mlhp.scalarField(3, 1.0), ifield)
        mlhp.integrateOnSurface(basis, supportIntegrand0, [res], surfaceQuadratures[reactionFace])
        support.append([i for i, r in enumerate(res) if r != 0.0])
        mlhp.fill(res, 0.0)

    residualIntegrand = mlhp.staticDomainIntegrand(kinematics, material, solution, computeTangent=False)

    mlhp.integrateOnDomain(basis, residualIntegrand, [res], domainQuadrature)

    reactionForce1 = [sum([res[i] * one[i] for i in ids]) for ids in support]

    # Reaction force computation method 2 (evaluate surface integral)
    reactionIntegrand = mlhp.reactionForceIntegrand(kinematics, material, solution)
    reactionForce2 = [mlhp.ScalarDouble(0.0) for _ in range(3)]
    
    mlhp.integrateOnSurface(basis, reactionIntegrand, reactionForce2, surfaceQuadratures[reactionFace])

    # Write output files
    postprocessors = [mlhp.solutionProcessor(3, solution), 
                      mlhp.functionProcessor(domain, "Domain"), 
                      mlhp.vonMisesProcessor(solution, kinematics, material)] 
                      
    postmesh0 = mlhp.gridOnCells([1, 1, 1], mlhp.PostprocessTopologies.Volumes)
    postmesh1 = mlhp.marchingCubesBoundary(resolution=[16] * 3, function=domain)

    mlhp.writeBasisOutput(basis, postmesh0, processors=postprocessors, writer=mlhp.PVtuOutput("outputs/immersed_beam_mesh"))
    mlhp.writeBasisOutput(basis, postmesh1, processors=postprocessors, writer=mlhp.PVtuOutput("outputs/immersed_beam_boundary"))

    return reactionForce1, [v.get() for v in reactionForce2]


class ImmersedBeamTest(unittest.TestCase):
    def test_beam(self):
        computedForce1, computedForce2 = computeImmersedBeam()
   
        #print(computedForce1, computedForce2)

        expectedForce1 = [0.9188900472824151, 0.051292405920671726, -0.3934466008426754]
        expectedForce2 = [0.9088664395517154, 0.0897414937122639, -0.40792079443694496]

        for expected1, computed1 in zip(expectedForce1, computedForce1):
            self.assertAlmostEqual(expected1, computed1, places=6)
    
        for expected2, computed2 in zip(expectedForce2, computedForce2):
            self.assertAlmostEqual(expected2, computed2)
    



import mlhp

D = 3

print("1. Setting up mesh and basis", flush=True)

# Setup triangulation domain
triangulation = mlhp.readStl("../csg.stl")
kdtree = mlhp.buildKdTree(triangulation)
domain = mlhp.implicitTriangulation(triangulation, kdtree)

# Setup discretization
youngsModulus = 200 * 1e9
poissonsRatio = 0.3

polynomialDegree = 1
nelements = [50] * D
alphaFCM = 1e-5
penalty = 1e5 * youngsModulus

origin, max = triangulation.boundingBox()

# ##################### FIX ####################
origin = [o - 1e-10 for o in origin]
max = [m + 1e-10 for m in max]

lengths = [m - o for o, m in zip(origin, max)]

grid = mlhp.makeRefinedGrid(nelements, lengths, origin)
basis = mlhp.makeHpTensorSpace(grid, polynomialDegree, nfields=D)

print(basis)

print("2. Allocating linear system", flush=True)

matrix = mlhp.allocateSparseMatrix(basis)
vector = mlhp.allocateRhsVector(matrix)

print("2. Computing weak boundary integrals", flush=True)

def createBoundaryQuadrature(func):
    filtered = mlhp.filterTriangulation(triangulation, mlhp.implicitFunction(D, func))
    intersected, celldata = mlhp.intersectTriangulationWithMesh(grid, filtered)
    quadrature = mlhp.triangulationQuadrature(intersected, celldata, polynomialDegree + 1)
    
    return intersected, celldata, quadrature
    
intersected0, celldata0, quadrature0 = createBoundaryQuadrature(f"x < {origin[0] + 1e-6}")
intersected1, celldata1, quadrature1 = createBoundaryQuadrature(f"x > {origin[0] + lengths[0] - 1e-6}")

integrand0 = mlhp.l2BoundaryIntegrand(mlhp.vectorField(D, [penalty] * D),
                                      mlhp.vectorField(D, [0.0] * D))

integrand1 = mlhp.neumannIntegrand(mlhp.vectorField(D, [1e3, 0.0, 0.0]))

mlhp.integrateOnSurface(basis, integrand0, [matrix, vector], quadrature0)
mlhp.integrateOnSurface(basis, integrand1, [vector], quadrature1)

print("4. Computing domain integral", flush=True)

E = mlhp.scalarField(D, 200 * 1e9)
nu = mlhp.scalarField(D, 0.3)
rhs = mlhp.vectorField(D, [0.0, 0.0, 0.0])

kinematics = mlhp.smallStrainKinematics(D) 
constitutive = mlhp.isotropicElasticMaterial(E, nu)
integrand = mlhp.staticDomainIntegrand(kinematics, constitutive, rhs)

quadrature = mlhp.momentFittingQuadrature(domain, 
    depth=polynomialDegree, epsilon=alphaFCM)

mlhp.integrateOnDomain(basis, integrand, [matrix, vector], quadrature=quadrature)

print("6. Solving linear system", flush=True)

#P = mlhp.additiveSchwarzPreconditioner(matrix, basis, dirichlet[0])
P = mlhp.diagonalPreconditioner(matrix)

dofs, norms = mlhp.cg(matrix, vector, M=P, maxiter=2000, residualNorms=True)

#print(f"cond K after domain integral: {numpy.linalg.cond(matrix.todense())}")
#import matplotlib.pyplot as plt
#plt.loglog(norms)
#plt.show()

print("7. Postprocessing solution", flush=True)

# Output solution on FCM mesh and boundary surface
gradient = mlhp.projectGradient(basis, dofs, quadrature)

processors = [mlhp.solutionProcessor(D, dofs, "Displacement"),
              mlhp.stressProcessor(gradient, kinematics, constitutive),
              mlhp.vonMisesProcessor(dofs, kinematics, constitutive, "VonMises1"),
              mlhp.vonMisesProcessor(gradient, kinematics, constitutive, "VonMises2"),
              mlhp.strainEnergyProcessor(gradient, kinematics, constitutive)]

intersected, celldata = mlhp.intersectTriangulationWithMesh(grid, triangulation, kdtree)

surfmesh = mlhp.associatedTrianglesCellMesh(intersected, celldata)

writer0 = mlhp.PVtuOutput(filename="outputs/linear_elasticity_fcm_stl_boundary")
writer1 = mlhp.PVtuOutput(filename="outputs/linear_elasticity_fcm_stl_fcmmesh")

mlhp.writeBasisOutput(basis, surfmesh, writer0, processors)
mlhp.writeBasisOutput(basis, writer=writer1, processors=processors)

# Output boundary surfaces
surfmesh0 = mlhp.associatedTrianglesCellMesh(intersected0, celldata0)
surfmesh1 = mlhp.associatedTrianglesCellMesh(intersected1, celldata1)

surfwriter0 = mlhp.VtuOutput(filename="outputs/linear_elasticity_fcm_stl_boundary0")
surfwriter1 = mlhp.VtuOutput(filename="outputs/linear_elasticity_fcm_stl_boundary1")
            
mlhp.writeMeshOutput(grid, surfmesh0, surfwriter0, [])
mlhp.writeMeshOutput(grid, surfmesh1, surfwriter1, [])



# This file is part of the mlhp project. License: See LICENSE
import mlhp
import numpy

from numba import cfunc, types, carray

targetTypes = [mlhp.AssemblyType.UnsymmetricMatrix, mlhp.AssemblyType.Vector]
ndim, nfields, ntargets = 3, 1, len(targetTypes)

print("Compiling...", flush=True)

# Use numba to compile element linear system callback into c function and pass function pointer to mlhp
@cfunc(types.float64(types.CPointer(types.float64)))
def rhs(xyzPtr):
    xyz = carray(xyzPtr, ndim)
    return numpy.sqrt(xyz[0] + xyz[1] + xyz[2])

@cfunc(types.float64(types.CPointer(types.float64)))
def conductivity(xyzPtr): return 1

@cfunc(types.void(types.CPointer(types.CPointer(types.float64)), # targets
                  types.CPointer(types.CPointer(types.float64)), # shapes
                  types.CPointer(types.CPointer(types.float64)), # mapping
                  types.CPointer(types.float64),                 # local coordinates
                  types.CPointer(types.float64),                 # history variables
                  types.CPointer(types.float64),                 # temporary storage
                  types.CPointer(types.int64),                   # location map
                  types.CPointer(types.int64),                   # total sizes
                  types.CPointer(types.int64),                   # field sizes
                  types.float64, types.float64, types.int64))    # det(J), weight, element index
def evaluate(targets_, shapes_, mapping_, rst_, history_, tmp_, locationMap_, sizes_, fieldSizes_, detJ, weight, ielement):
    ndof, ndofpadded = carray(sizes_, 2)
    xyz_ = carray(mapping_, 1)[0]
    
    Ke = carray(carray(targets_, ntargets)[0], (ndof, ndofpadded))
    Fe = carray(carray(targets_, ntargets)[1], ndofpadded)
    
    N = carray(carray(shapes_, nfields)[0], (4, ndofpadded))[0]
    B = carray(carray(shapes_, nfields)[0], (4, ndofpadded))[1:]
    
    k = conductivity(xyz_) * weight * detJ
    f = rhs(xyz_) * weight * detJ
    
    for idof in range(ndof):
        for jdof in range(ndof):
            for axis in range(ndim):
                Ke[idof, jdof] += B[axis, idof] * B[axis, jdof] * k
    
    for idof in range(ndof):
        Fe[idof] += N[idof] * f

print("Preprocessing...", flush=True)

integrand = mlhp.domainIntegrand(ndim, evaluate, types=targetTypes, maxdiff=1)

# Preprocess
mesh = mlhp.makeRefinedGrid([100, 100, 100])
basis = mlhp.makeHpTrunkSpace(mesh, degrees=1, nfields=nfields)

dirichlet = mlhp.integrateDirichletDofs(mlhp.scalarField(ndim, 0.0), basis, [1, 3, 5])

print(mesh)
print(basis)

print("Assembling...", flush=True)

matrix = mlhp.allocateSparseMatrix(basis, dirichlet[0])
vector = mlhp.allocateRhsVector(matrix)

print(matrix, flush=True)

mlhp.integrateOnDomain(basis, integrand, [matrix, vector], dirichletDofs=dirichlet)

print("Solving...", flush=True)

dofs = mlhp.inflateDofs(mlhp.makeCGSolver()(matrix, vector), dirichlet)

print("Postprocessing...", flush=True)

processors = [mlhp.solutionProcessor(ndim, dofs, "Temperature" )]

sampling = mlhp.gridOnCells(mlhp.degreeOffsetResolution(basis), mlhp.PostprocessTopologies.Volumes)
output = mlhp.PVtuOutput(filename="outputs/example_poisson_compiled")
            
mlhp.writeBasisOutput(basis, sampling, output, processors)

print("Done.")

# # Alternatively: use cffi to directly compile C code
# import cffi
# ffibuilder = cffi.FFI()
# ffibuilder.cdef("""extern unsigned long long address;""")
# ffibuilder.set_source("_integrand_cffi",
#     """
#         #include "math.h"
#         
#         double rhs(double* xyz)
#         {
#             return sqrt(xyz[0] + xyz[1] + xyz[2]);
#         }
#         
#         double conductivity(double* xyz)
#         {
#             return 1.0;
#         }
#         
#         void integrand(double** targets,     double** shapes,     double** mapping, 
#                        double*  rst,         double*  history,    double*  tmp, 
#                        int64_t* locationMap, int64_t* totalSizes, int64_t* fieldSizes, 
#                        double   detJ,        double   weight,     int64_t  ielement)
#         {
#            int64_t ndof = totalSizes[0];
#            int64_t ndofpadded = totalSizes[1];
#            
#            double* N = shapes[0];
#            double* B = shapes[0] + ndofpadded;
#            
#            double k = conductivity(mapping[0]) * weight * detJ;
#            double f = rhs(mapping[0]) * weight * detJ;
#            
#            for(int64_t idof = 0; idof < ndof; ++idof)
#            {
#                for(int64_t jdof = 0; jdof < ndof; ++jdof)
#                {
#                    for(int64_t axis = 0; axis < 3; ++axis)
#                    {
#                        targets[0][idof * ndofpadded + jdof] += B[axis * ndofpadded + idof] * 
#                                                                B[axis * ndofpadded + jdof] * k;
#                    }
#                }
#                
#                targets[1][idof] += N[idof] * f;
#            }
#         }
#         
#         unsigned long long address = (unsigned long long)&integrand;
#     """)
# ffibuilder.compile()
# from _integrand_cffi import lib as evaluate



import mlhp

D = 3

print("1. Setting up mesh and basis", flush=True)

refinementDepth = 0
polynomialDegree = 2
nelements = [10] * D
lengths = [1.0] * D

alphaFCM = 1e-3 # needs better preconditioning

domain = mlhp.invert(mlhp.implicitCube([0.0,0.101,0.101], [1.0, 0.899, 0.899]))

strategy = mlhp.refineTowardsBoundary(domain, refinementDepth)

grid = mlhp.makeRefinedGrid(nelements, lengths)
grid.refine(strategy)

basis = mlhp.makeHpTensorSpace(grid, polynomialDegree, nfields=D)

print("2. Computing dirichlet boundary conditions", flush=True)

dirichlet = mlhp.integrateDirichletDofs(mlhp.vectorField(D, [0.0] * D), basis, [0])

print("3. Setting up physics", flush=True)

E = mlhp.scalarField(D, 200 * 1e9)
nu = mlhp.scalarField(D, 0.3)
rhs = mlhp.vectorField(D, [0.0, 0.0, 78.5 * 1e3])

kinematics = mlhp.smallStrainKinematics(D) 
constitutive = mlhp.isotropicElasticMaterial(E, nu)
integrand = mlhp.staticDomainIntegrand(kinematics, constitutive, rhs)

print("4. Allocating linear system", flush=True)

matrix = mlhp.allocateSparseMatrix(basis, dirichlet[0])
vector = mlhp.allocateRhsVector(matrix)

print("5. Integrating linear system", flush=True)

quadrature = mlhp.momentFittingQuadrature(domain, 
    depth=polynomialDegree + 1, epsilon=alphaFCM)

mlhp.integrateOnDomain(basis, integrand, [matrix, vector], 
    dirichletDofs=dirichlet, quadrature=quadrature)

print("6. Solving linear system", flush=True)

P = mlhp.additiveSchwarzPreconditioner(matrix, basis, dirichlet[0])
#P = mlhp.diagonalPreconditioner(matrix)

interiorDofs, norms = mlhp.cg(matrix, vector, M=P, maxiter=1000, residualNorms=True)

allDofs = mlhp.inflateDofs(interiorDofs, dirichlet)

print("7. Postprocessing solution", flush=True)

processors = [mlhp.solutionProcessor(D, allDofs, "Displacement"),
              mlhp.functionProcessor(domain)]

postmesh = mlhp.gridOnCells([polynomialDegree + 3] * D)
writer = mlhp.PVtuOutput(filename="outputs/linear_elasticity")
            
mlhp.writeBasisOutput(basis, postmesh, writer, processors)



$ cat printMedians.py
#!/usr/bin/python

import numpy
import performance_benchmark_data

def printMatrix( M ):
    for row in M:
        string = str( row[0] )
        for entry in row[1:]:
            string += ", " + str( entry )
        print( string )
    print( "" )

def printResult( data, name ):
    print( "--- " + str( name ) + " time ---" )
    for iAlign in range( data.shape[0] ):
        printMatrix( numpy.median( data[iAlign], axis=-1 ) )

data = numpy.array( performance_benchmark_data.data )

print( str( data.shape[0] ) + " data sets." )
print( str( data.shape[1] ) + " alignments." )
print( str( data.shape[4] ) + " number of runs." )

printResult( data[0], "Grid" )
printResult( data[1], "Basis" )
printResult( data[2], "Allcoation" )
printResult( data[3], "Shapes" )
printResult( data[4], "Matrix" )
printResult( data[5], "Scatter" )
printResult( data[6], "Total assembly" )
printResult( data[7], "Total" )
